/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.xwiki.notifications;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.apache.commons.lang3.builder.HashCodeBuilder;
import org.xwiki.eventstream.DocumentEventType;
import org.xwiki.eventstream.Event;
import org.xwiki.model.reference.DocumentReference;
import org.xwiki.stability.Unstable;
import org.xwiki.text.StringUtils;

/**
 * A group of similar events that compose a "composite" event.
 *
 * @version $Id: 853b47b5dda9d3a0687c904f43ef88e70c4f949c $
 * @since 9.4RC1
 */
public class CompositeEvent
{
    private List<Event> events = new ArrayList<>();

    private int similarityBetweenEvents;

    /**
     * Construct a CompositeEvent.
     * 
     * @param event the first event of the composite event
     */
    public CompositeEvent(Event event)
    {
        events.add(event);
    }

    /**
     * Copy constructor of {@link CompositeEvent}.
     *
     * @param compositeEvent the composite event to copy
     */
    public CompositeEvent(CompositeEvent compositeEvent)
    {
        this.events = new ArrayList<>(compositeEvent.events);
        this.similarityBetweenEvents = compositeEvent.similarityBetweenEvents;
    }

    /**
     * @return the list of the ids of the events
     */
    public List<String> getEventIds()
    {
        ArrayList results = new ArrayList(events.size());
        for (Event event : events) {
            results.add(event.getId());
        }
        return results;
    }

    /**
     * Add an event to the composite event.
     * 
     * @param event the event to add
     * @param similarity the similarity between the event to add and the events of the composite events
     * @throws NotificationException if the addition is illegal (lower similarity for example)
     * @deprecated now that we have a {@link GroupingEventStrategy} this method has no real reason to be used,
     * {@link #add(Event)} should be used instead.
     */
    @Deprecated(since = "15.5RC1")
    public void add(Event event, int similarity) throws NotificationException
    {
        if (similarity < similarityBetweenEvents) {
            throw new NotificationException("Invalid addition of an event inside a CompositeEvent");
        }
        similarityBetweenEvents = similarity;
        this.add(event);
    }

    /**
     * Add an event to the composite event and order it based on the dates.
     *
     * @param event the event to add
     * @since 15.5RC1
     */
    @Unstable
    public void add(Event event)
    {
        events.add(event);
        // Ensure the events are always sorted by date (more recent firsts)
        Collections.sort(events, (e1, e2) -> e2.getDate().compareTo(e1.getDate()));
    }

    /**
     * @return the greatest similarity between events of the composite event
     * @deprecated this method shouldn't be used anymore now we have a {@link GroupingEventStrategy}
     */
    @Deprecated(since = "15.5RC1")
    public int getSimilarityBetweenEvents()
    {
        return similarityBetweenEvents;
    }

    /**
     * @return the events that compose the current object
     */
    public List<Event> getEvents()
    {
        return events;
    }

    /**
     * @return the type of the events that compose the current object
     */
    public String getType()
    {
        // Fallback type
        String type = events.get(0).getType();
        for (Event event : events) {
            // We are most interested in "advanced" event that we are in "core" events such as "create" or "update",
            // which often are the technical consequences of the real event (ex: a comment has been added).
            // FIXME: this sounds really like a hack: we should probably instead allow to set the type.
            if (StringUtils.isNotBlank(event.getType()) && !DocumentEventType.CREATE.equals(event.getType())
                && !DocumentEventType.UPDATE.equals(event.getType())) {
                type = event.getType();
            }
        }
        return type;
    }

    /**
     * @return the groupId of the first event of the current object
     */
    // FIXME: To deprecate? Not so sure what it's used for...
    public String getGroupId()
    {
        return events.get(0).getGroupId();
    }

    /**
     * @return the document of the first event of the current object
     */
    // FIXME: We probably need a getDocuments
    public DocumentReference getDocument()
    {
        return events.get(0).getDocument();
    }

    /**
     * @return the users who performed the events
     */
    public Set<DocumentReference> getUsers()
    {
        // Use a linked hash set to follow the order in which events were added:
        // this avoids problems when testing
        Set<DocumentReference> users = new LinkedHashSet<>();
        for (Event event : events) {
            users.add(event.getUser());
        }
        return users;
    }

    /**
     * @return the date of the most recent event
     * @since 9.11RC1
     */
    public Date getDate()
    {
        Iterator<Event> it = events.iterator();
        Event chosenEvent = it.next();
        while (it.hasNext()) {
            Event event = it.next();
            if (event.getDate().compareTo(chosenEvent.getDate()) > 0) {
                chosenEvent = event;
            }
        }
        return chosenEvent.getDate();
    }

    /**
     * @return the dates of the events, sorted by descending order
     */
    public List<Date> getDates()
    {
        List<Date> dates = new ArrayList<>();
        for (Event event : events) {
            dates.add(event.getDate());
        }
        Collections.sort(dates, Collections.reverseOrder());
        return dates;
    }

    /**
     * Remove an event from the current object.
     * 
     * @param event the event to remove
     */
    public void remove(Event event)
    {
        events.remove(event);
    }

    @Override
    public boolean equals(Object obj)
    {
        if (obj == this) {
            return true;
        }

        if (obj instanceof CompositeEvent) {
            CompositeEvent compositeEvent = (CompositeEvent) obj;

            EqualsBuilder builder = new EqualsBuilder();
            builder.append(getSimilarityBetweenEvents(), compositeEvent.getSimilarityBetweenEvents());
            builder.append(getEvents(), compositeEvent.getEvents());

            return builder.build();
        }

        return false;
    }

    @Override
    public int hashCode()
    {
        HashCodeBuilder builder = new HashCodeBuilder();

        builder.append(getSimilarityBetweenEvents());
        builder.append(getEvents());

        return builder.build();
    }
}
