/*
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package com.xpn.xwiki.store;

import java.util.Date;

import javax.inject.Named;
import javax.inject.Singleton;

import org.suigeneris.jrcs.rcs.Archive;
import org.suigeneris.jrcs.rcs.Version;
import org.xwiki.component.annotation.Component;

import com.xpn.xwiki.XWikiContext;
import com.xpn.xwiki.XWikiException;
import com.xpn.xwiki.doc.XWikiAttachment;
import com.xpn.xwiki.doc.XWikiAttachmentArchive;

/**
 * Void store for attachment versioning when it is disabled. ("xwiki.store.attachment.versioning=0" parameter is set in
 * xwiki.cfg) It says what there is only one version of attachment - latest. It doesn't store anything. It is safe to
 * use with any stores.
 *
 * @version $Id: 971adb2818c92d15f1b50f8d580b686e4f60e6c4 $
 * @since 1.4M2
 */
@Component
@Named("void")
@Singleton
public class VoidAttachmentVersioningStore implements AttachmentVersioningStore
{
    /**
     * Constructor used by {@link com.xpn.xwiki.XWiki} during storage initialization.
     *
     * @param context The current context.
     * @deprecated 1.6M1. Use ComponentManager.lookup(AttachmentVersioningStore.class) instead.
     */
    @Deprecated
    public VoidAttachmentVersioningStore(XWikiContext context)
    {
    }

    /**
     * Empty constructor needed for component manager.
     */
    public VoidAttachmentVersioningStore()
    {
    }

    @Override
    public String getHint()
    {
        return "void";
    }

    @Override
    public void deleteArchive(XWikiAttachment attachment, XWikiContext context, boolean transaction)
        throws XWikiException
    {
        // Don't do anything since it's a void implementation.
    }

    @Override
    public void saveArchive(XWikiAttachmentArchive archive, XWikiContext context, boolean transaction)
        throws XWikiException
    {
        // Don't do anything since it's a void implementation.
    }

    @Override
    public XWikiAttachmentArchive loadArchive(XWikiAttachment attachment, XWikiContext context, boolean transaction)
        throws XWikiException
    {
        XWikiAttachmentArchive archive = attachment.getAttachment_archive();
        if (!(archive instanceof VoidAttachmentArchive)) {
            archive = new VoidAttachmentArchive(attachment);
        }
        attachment.setAttachment_archive(archive);
        return archive;
    }

    /**
     * Void realization of AttachmentArchive. It says what there is only one version of attachment - latest. Class is
     * public because used in super.clone() via getClass().newInstance()
     */
    public static class VoidAttachmentArchive extends XWikiAttachmentArchive
    {
        /**
         * Default constructor. Used in super.clone().
         */
        public VoidAttachmentArchive()
        {
        }

        /**
         * Helper constructor.
         *
         * @param attachment attachment of this archive
         */
        public VoidAttachmentArchive(XWikiAttachment attachment)
        {
            setAttachment(attachment);
        }

        @Override
        public void updateArchive(final byte[] data, final XWikiContext context) throws XWikiException
        {
            updateArchive(context);
        }

        @Override
        public void updateArchive(XWikiContext context) throws XWikiException
        {
            getAttachment().incrementVersion();
            getAttachment().setDate(new Date());
        }

        @Override
        public void setArchive(byte[] data) throws XWikiException
        {
            // Don't do anything since it's a void implementation.
        }

        @Override
        public byte[] getArchive(XWikiContext context) throws XWikiException
        {
            return new byte[0];
        }

        @Override
        public void setRCSArchive(Archive archive)
        {
            // Don't do anything since it's a void implementation.
        }

        @Override
        public Version[] getVersions()
        {
            return new Version[] { getAttachment().getRCSVersion() };
        }

        @Override
        public XWikiAttachment getRevision(XWikiAttachment attachment, String rev, XWikiContext context)
            throws XWikiException
        {
            return (attachment.getVersion().equals(rev)) ? attachment : null;
        }

        @Override
        public Object clone()
        {
            // super.clone() is needed for checkstyle
            super.clone();
            return new VoidAttachmentArchive(getAttachment());
        }
    }
}
